import logging
import inspect
from functools import wraps
from contextlib import contextmanager, ExitStack
from typing import Callable, Any, List, Dict, Tuple
from file_or_name.utils import parameterize, get_first_parameter, ShadowPage


__all__ = ["file_or_name"]
LOGGER = logging.getLogger("file_or_name")
UTF_8 = "utf-8"


@contextmanager
def open_files(files: Dict[str, str], function: Callable, *args: List[Any], **kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Open files that are arguments to this function call and close ones we opened after the call.

    :param files: A mapping of argument name to file mode. This should include all arguments that are files
    :param function: The user defined function we are opening these files for
    :param args: A list of positional args function was called with
    :param kwargs: A dict of keyword args function was called with
    :returns: A dict mapping argument names to argument values with the opened files inserted.
        function should now be called with this dict ** unpacked.
    """
    call_args = inspect.getcallargs(function, *args, **kwargs)
    with ExitStack() as stack:
        for file_name, mode in files.items():
            LOGGER.debug("Opening file %s in mode %s", file_name, mode)
            if file_name not in call_args:
                raise ValueError(f"Argument {file_name} is missing and expected to be opened in {mode} mode.")
            if isinstance(call_args[file_name], str):
                if mode.startswith("s"):
                    if "w" not in mode:
                        raise ValueError(
                            f"Shadow paging (requested by pre-pending `s` to the {call_args[file_name]}'s file mode) only supported for writing, got {mode[1:]}"
                        )
                    call_args[file_name] = stack.enter_context(
                        ShadowPage(call_args[file_name], mode[1:], encoding=None if "b" in mode else UTF_8)
                    )
                else:
                    call_args[file_name] = stack.enter_context(
                        open(call_args[file_name], mode=mode, encoding=None if "b" in mode else UTF_8)
                    )
        yield call_args


@parameterize
def file_or_name(function: Callable, **kwargs: Dict[str, str]) -> Callable:
    """Transparent allow arguments to be either strings or open files.

    kwargs are of the form name=mode, This is used to create a mapping of arguments
    to the mode that file should be opened with, for example if the argument named
    wf should be opened in write mode you would call
        `@file_or_name(wf='w')`

    If there are no kwargs it is assumed the first argument is a file that will be
    opened in read mode "r"

    If you need a file to be opened with extra arguments, for example `newline=''`
    for files the stdlib csv writer, you should manually open the file and pass
    the resulting file object in.
    """
    files = kwargs
    # If no file modes are specified in the kwargs we set the first argument to be opened in read mode
    if not files:
        first = get_first_parameter(function)
        LOGGER.debug("No file parameters provided, using %s='r'", first)
        files[first] = "r"
    # We need to check if we are a generator out here because if we waited to check in the
    # open_arg_files function then open_arg_files would be a generator no matter what because
    # it would have a yield in it (even if that code path wasn't executed for a function)
    if inspect.isgeneratorfunction(function):

        @wraps(function)
        def open_arg_files(*args, **kwargs):
            with open_files(files, function, *args, **kwargs) as call_args:
                yield from function(**call_args)

    else:

        @wraps(function)
        def open_arg_files(*args, **kwargs):
            with open_files(files, function, *args, **kwargs) as call_args:
                return function(**call_args)

    return open_arg_files



import os
import logging
import inspect
from functools import wraps
from typing import Callable, Optional
from tempfile import NamedTemporaryFile


__all__ = ["parameterize", "get_first_parameter", "ShadowPage"]
LOGGER = logging.getLogger("file_or_name")


def parameterize(function: Callable) -> Callable:
    """Allow a decorator to be called without parentheses if no kwargs are given.

    parameterize is a decorator, function is also a decorator.
    """

    @wraps(function)
    def decorator(*args, **kwargs):
        """If a decorator is called with only the wrapping function just execute the real decorator.
           Otherwise return a lambda that has the args and kwargs partially applied and read to take a function as an argument.

        *args, **kwargs are the arguments that the decorator we are parameterizing is called with.

        the first argument of *args is the actual function that will be wrapped
        """
        if len(args) == 1 and not kwargs and callable(args[0]):
            return function(args[0])
        return lambda wrappee: function(wrappee, *args, **kwargs)

    return decorator


def get_first_parameter(function: Callable) -> str:
    """Get the name of the first parameter of a function.

    :param function: The function we want the name of the first parameter of
    :returns: The name of the first parameters
    """
    sig = inspect.signature(function)
    return list(sig.parameters.keys())[0]


class ShadowPage:
    """Store updates to a different copy of the output file and swing pointers for an atomic write."""

    def __init__(self, path: str, mode: str = "wb", dir: Optional[str] = None, encoding: Optional[str] = None):
        self.path = path
        # Don't try to delete the temp file when cleaning up, it will either be removed in the
        # swing to the real file or it should be left for debugging
        self.temp_file = NamedTemporaryFile(mode=mode, delete=False, dir=dir, encoding=encoding)
        LOGGER.debug("Opening shadow file for %s at %s", self.path, self.temp_file.name)

    def __enter__(self):
        """Go the normal thing temp files do when entering a context block."""
        self.temp_file.__enter__()
        return self

    def write(self, *args, **kwargs):
        """Proxy writes to the temp file."""
        return self.temp_file.write(*args, **kwargs)

    def __exit__(self, exc, value, tb):
        # If we exited this context manager normally swing the pointer to the real path
        if exc is None:
            # Try to set the permission on the temp file to match the real file
            try:
                stat = os.stat(self.path)
                os.chown(self.temp_file.name, stat.st_uid, stat.st_gid)
                os.chmod(self.temp_file.name, stat.st_mode)
            except FileNotFoundError:
                pass
            os.replace(self.temp_file.name, self.path)
            LOGGER.debug("Replacing %s with shadow file %s", self.path, self.temp_file.name)
        # If we found an error don't try to replace the old file
        else:
            LOGGER.debug("Context closed do to %s rolling back update.", exc.__name__)
        # Normal tempfile cleanup
        self.temp_file.__exit__(exc, value, tb)
